functor TrmHelper(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    structure Trm: TRM
	    sharing Typ.Globals = Globals
	        and TypHelper.Typ = Typ
	        and TypHelper.TEnv = TEnv
	        and TEnv.Typ = Typ
		and Trm.TypHelper = TypHelper
	        and Globals.Id = Typ.Id
	        and Typ.Id = TEnv.Id
	        and TEnv.Id = TypHelper.Id
	    ) : TRMHELPER
	    = struct

structure Globals = Globals
structure Typ = Typ
structure TypHelper = TypHelper
structure TEnv = TEnv
structure Trm = Trm

open Globals
open Trm
open TypHelper
open TEnv
open Typ
open Pp

datatype PreTrm = 
	    PREVAR of Id.T
	  | PREABS of Id.T * TypHelper.PreTyp * PreTrm
	  | PREAPP of PreTrm * PreTrm
	  | PRETABS of Id.T * TypHelper.PreTyp * PreTrm
	  | PRETAPP of PreTrm * TypHelper.PreTyp
          | PRERCDV of (Id.T * PreTrm) list
	  | PREPROJ of PreTrm * Id.T 
	  | PREOPEN of Id.T * Id.T * PreTrm * PreTrm
	  | PREPACK of PreTrm * TypHelper.PreTyp * TypHelper.PreTyp 
	  | PREASSERT of PreTrm * TypHelper.PreTyp 
	  | PREABBREV of Id.T * TypHelper.PreTyp * PreTrm 
	  | PRELET of Id.T * PreTrm * PreTrm 
	  | PREPRIM of string * TypHelper.PreTyp

exception IllegalPrimitive of PreTrm

fun make_prim s = s

val Int = Id.intern "Int"

fun debruijnify te (PREVAR i) = 
      VAR(index te i)
  | debruijnify te (PREABS(i,ptyp,ptrm)) =
      ABS(i, TypHelper.debruijnify te ptyp, 
      		    debruijnify (push_binding te i dummytype) ptrm)
  | debruijnify te (PREAPP(ptrm1,ptrm2)) =
      APP(debruijnify te ptrm1, debruijnify te ptrm2)
  | debruijnify te (PRETABS(i,ptyp,ptrm)) =
      TABS(i, TypHelper.debruijnify te ptyp, 
      		     debruijnify (push_bound te i dummytype) ptrm)
  | debruijnify te (PRETAPP(ptrm,ptyp)) =
      TAPP(debruijnify te ptrm, TypHelper.debruijnify te ptyp)
  | debruijnify te (PREPROJ(ptrm,n)) = 
      PROJ(debruijnify te ptrm, n) 
  | debruijnify te (PRERCDV(l)) = 
      RCDV(map (fn (n,ptrm) => (n,debruijnify te ptrm)) l)
  | debruijnify te (PREOPEN(a,x,e1,e2)) = 
      OPEN(a, x, debruijnify te e1, 
      	   debruijnify (push_binding (push_bound te a dummytype) x dummytype) 
      	   	       e2) 
  | debruijnify te (PREPACK(e,t,h)) = 
      PACK(debruijnify te e, TypHelper.debruijnify te t, 
      	   TypHelper.debruijnify te h) 
  | debruijnify te (PREASSERT(e,t)) = 
      ASSERT(debruijnify te e, TypHelper.debruijnify te t)
  | debruijnify te (PREABBREV(i,t,e)) = 
      ABBREV(i, TypHelper.debruijnify te t, 
                debruijnify (push_bound te i dummytype) e)
  | debruijnify te (PRELET(i,e1,e2)) = 
      LET(i, debruijnify te e1,
                debruijnify (push_binding te i dummytype) e2)
  | debruijnify te (PREPRIM(p,t)) = 
      PRIM(make_prim p, TypHelper.debruijnify te t)

fun prt_nested te e = let
  fun pe (VAR(i)) = 
      write (Id.tostr (lookup_name te i))
    | pe (ABS(n,t,e)) = 
      (write "(fun(";
       write (Id.tostr n);
       write ":";
       TypHelper.prt te t;
       write ") ";
       ubreak 2;
       prt_nested (TEnv.push_binding te n t) e;
       write ")")
    | pe (APP(e1,e2)) = 
      (write "(";
       prt_nested te e1;
       write " ";
       ubreak 0;
       prt_nested te e2;
       write ")")
    | pe (TABS(n,t,e)) = 
      (write "(fun(";
       write (Id.tostr n);
       write "<";
       TypHelper.prt te t;
       write ") ";
       ubreak 2;
       prt_nested (TEnv.push_bound te n t) e;
       write ")")
    | pe (TAPP(e1,t2)) = 
      (write "(";
       prt_nested te e1;
       ubreak ~1;
       write " ";
       TypHelper.prt te t2;
       write ")")
    | pe (RCDV(fields)) =
      (write "{";
       mapunit_tuple (fn (l,e) => (write (Id.tostr l);
                                   write "=";
                                   prt_nested te e))
                     (fn () => (write ", "; ubreak 1))
                     fields;
       write "}")
    | pe (PROJ(e,l)) =
      (prt_nested te e;
       write ".";
       write (Id.tostr l))
    | pe (OPEN(a,x,e,b)) =
      (write "open ";
       prt_nested te e;
       ubreak ~1;
       write " as <";
       write (Id.tostr a);
       write ",";
       write (Id.tostr x);
       write "> ";
       ubreak 0;
       write "in ";
       prt_nested (TEnv.push_binding
                      (TEnv.push_bound te a dummytype)
                      x dummytype)
                   b;
       write " end")
    | pe (PACK(e,t,h)) =
      (write "<";
       TypHelper.prt te h;
       write ", ";
       ubreak 1;
       prt_nested te e;
       write ": ";
       TypHelper.prt te t)
    | pe (ASSERT(e,t)) =
      (write "assert ";
       prt_nested te e;
       write " : ...")
    | pe (ABBREV(a,t,e)) =
      (write "let ";
       write (Id.tostr a);
       write " = ";
       TypHelper.prt te t;
       ubreak ~1;
       write " in ";
       prt_nested (TEnv.push_bound te a t) e;
       write " end")
    | pe (LET(x,e,b)) =
      (write "let ";
       write (Id.tostr x);
       write " = ";
       prt_nested te e;
       ubreak ~1;
       write " in ";
       prt_nested (TEnv.push_bound te x dummytype) b;
       write " end")
    | pe (PRIM(p,t)) =
      (write "<<"; 
       prt_primitive p;
       write " : ";
       TypHelper.prt te t;
       write ">>")
  in setb();
     pe e;
     endb()
  end

fun prt_long te e inner = let
  fun opn s = (if inner then write s else (); setb())
  fun cls s = (endb(); if inner then write s else ())
  fun pe (VAR(i)) = 
        write (Id.tostr (lookup_name te i))
    | pe (ABS(n,t,e)) = 
        (opn "("; write "fun("; write (Id.tostr n); write ":"; 
         TypHelper.prt te t; write ") "; ubreak 2; 
         prt_long (TEnv.push_binding te n t) e false; cls ")")
    | pe (APP(e1,e2)) = 
        (opn "("; prt_long te e1 true; write " "; ubreak 0;
         prt_long te e2 true; cls ")")
    | pe (TABS(n,t,e)) = 
        (opn "("; write "fun("; write (Id.tostr n); write "<"; 
         TypHelper.prt te t; write ") "; ubreak 2; 
         prt_long (TEnv.push_bound te n t) e false; cls ")")
    | pe (TAPP(e1,t2)) = 
        (opn "("; prt_long te e1 false; ubreak ~1; write " ";
         TypHelper.prt te t2; cls ")")
    | pe (RCDV(fields)) =
        (write "{"; setb();
         mapunit_tuple (fn (l,e) => (write (Id.tostr l);
                                     write "=";
                                     prt_long te e false))
                       (fn () => (write ", "; ubreak 0))
                       fields;
         endb(); write "}")
    | pe (PROJ(e,l)) =
        (prt_long te e true; write "."; write (Id.tostr l))
    | pe (OPEN(a,x,e,b)) =
        (setb(); write "open "; prt_long te e false; ubreak ~1; write " as <";
         write (Id.tostr a); write ","; write (Id.tostr x); write "> "; 
         ubreak 0; write "in ";
         prt_long (TEnv.push_binding
                        (TEnv.push_bound te a dummytype)
                        x dummytype)
                     b false;
         write " end"; endb())
    | pe (PACK(e,t,h)) =
        (setb(); write "<"; TypHelper.prt te h; write ", ";
         ubreak 1; prt_long te e; 
         ubreak 0; write ">: "; TypHelper.prt te t; endb())
    | pe (ASSERT(e,t)) =
        (opn "("; write "assert "; prt_long te e false; ubreak 2;
         write " : ..."; cls ")")
    | pe (ABBREV(a,t,e)) =
        (setb(); write "let "; write (Id.tostr a); write " = ";
         TypHelper.prt te t; ubreak ~1; write " in "; 
         prt_long (TEnv.push_bound te a t) e; write " end"; endb())
    | pe (LET(x,e,b)) =
        (setb(); write "let "; write (Id.tostr x); write " = "; 
         prt_long te e false; ubreak ~1; write " in "; 
         prt_long (TEnv.push_bound te x dummytype) b false;
         write " end"; endb())
    | pe (PRIM(p,t)) =
        (write "<<"; prt_primitive p; write " : "; TypHelper.prt te t;
         write ">>")
  in pe e
  end

fun prt_short te e = let
  fun pe (VAR(i)) = write (Id.tostr (lookup_name te i))
    | pe (ABS(n,t,e)) = write "<fn>"
    | pe (APP(e1,e2)) = write "<app>"
    | pe (TABS(n,t,e)) = write "<tfn>"
    | pe (TAPP(e1,t2)) = write "<tapp>"
    | pe (RCDV(fields)) = 
      (write "{";
       mapunit_tuple (fn (l,e) => (write (Id.tostr l);
                                   write "=";
                                   pe e))
                     (fn () => write ", ")
                     fields;
       write "}")
    | pe (PROJ(e,l)) = 
      (pe e;
       write ".";
       write (Id.tostr l))
    | pe (OPEN(a,x,e,b)) = write "<open_exp>"
    | pe (PACK(e,t,h)) = write "<pack_exp>"
    | pe (ASSERT(e,t)) = write "<assertion>"
    | pe (ABBREV(a,t,e)) = write "<type_abbrev>"
    | pe (LET(x,e,b)) = write "<let_exp>"
    | pe (PRIM(p,t)) = prt_primitive p
  in pe e
  end

fun prt te trm = prt_long te trm false

end

